package de.hftstuttgart.unifiedticketing.core;

import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * <b>Ticket representation</b><br/>
 * <br/>
 * This class is the generic representation of a ticket for the supported systems.
 * It defines a few fields as well as getters and setters for them.<br/>
 * <br/>
 * Next there's a save method, to persist your changes back into the real ticket system.
 * This is is a method implemented in each of the system specific child classes.
 *
 * @param <TS> Ticketsystem a ticket instance belongs to
 * @param <T> child implementation of this class an instance belongs to
 */
public abstract class Ticket<TS extends TicketSystem, T extends Ticket>
{
    private static Logger logger = Logging.getLogger(Ticket.class.getName());

    protected Set<TicketAssignee> assignees;
    protected String description;
    protected String id;
    protected Set<String> labels;
    protected boolean open;
    protected String title;

    protected Set<String> updatedFields;

    protected TS parent;

    protected enum FieldNames
    {
        ASSIGNEES,
        DESCRIPTION,
        ID,
        LABELS,
        OPEN,
        TITLE,
    }

    /**
     * @param parent Ticketsystem instance, this ticket is from
     */
    protected Ticket(TS parent)
    {
        if (parent == null)
        {
            logger.log(Level.SEVERE, "every ticket instance needs a belonging ticket system, but received null!");
            throw new AssertionException(
                String.format("Got null as parent ticket system on creating new %s", this.getClass().getSimpleName()));
        }

        this.assignees = new HashSet<>();
        this.labels = new HashSet<>();
        this.open = true;
        this.parent = parent;
        this.updatedFields = new HashSet<>();
    }

    public Set<? extends TicketAssignee> getAssignees()
    {
        if (this.assignees == null) return new HashSet<>(); // TODO remove as assignees should NEVER turn null
        else return this.assignees;
    }

    /**
     * adds an new assignee to the ticket, by the systems api primary identifier for assignees
     * @return this ticket
     */
    public abstract T addAssignee(String identifier);

    /**
     * clears all assignees
     * @return this ticket
     */
    public T clearAssignees()
    {
        this.assignees.clear();
        this.updatedFields.add(FieldNames.ASSIGNEES.name());
        return (T) this;
    }

    /**
     * removes a single assignee identified by the platforms api primary identifier for assignees
     * @return this ticket
     */
    public abstract T removeAssignee(String identifier);

    public String getDescription()
    {
        return description;
    }

    /**
     * replaces the current ticket description with the new given one
     * @return this ticket
     */
    public T setDescription(String description)
    {
        this.description = description;
        this.updatedFields.add(FieldNames.DESCRIPTION.name());

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.DESCRIPTION.name()
        ));

        return (T) this;
    }

    public String getId()
    {
        return id;
    }

    /**
     * Adds the given label, preserving the yet assigned ones
     * @return this ticket
     */
    public T addLabel(String label)
    {
        this.labels.add(label);
        this.updatedFields.add(FieldNames.LABELS.name());

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.LABELS.name()
        ));

        return (T) this;
    }

    public Set<String> getLabels()
    {
        return labels;
    }

    /**
     * replaces all labels of the ticket with the new ones
     * @param labels collection to replace old ones
     * @return this ticket
     */
    public T setLabels(Set<String> labels)
    {
        this.labels = labels;
        this.updatedFields.add(FieldNames.LABELS.name());

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.LABELS.name()
        ));

        return (T) this;
    }

    /**
     * replaces all labels of the ticket with the new ones
     * @param labels array to replace old ones
     * @return this ticket
     */
    public T setLabels(String[] labels)
    {
        logger.log(Level.FINEST, String.format(
            "[Ticket %s] transforming labels Array to Set",
            this.id
        ));
        return this.setLabels(new HashSet<>(Arrays.asList(labels)));
    }

    public boolean isOpen()
    {
        return open;
    }

    /**
     * set's the ticket state to a open considered state
     * @return this ticket
     */
    public T open()
    {
        this.updatedFields.add(FieldNames.OPEN.name());
        this.open = true;

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.OPEN.name()
        ));

        return (T) this;
    }

    /**
     * set's the ticket state to a closed considered state
     * @return this ticket
     */
    public T close()
    {
        this.updatedFields.add(FieldNames.OPEN.name());
        this.open = false;

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.OPEN.name()
        ));


        return (T) this;
    }

    public TS getParent()
    {
        return parent;
    }

    public String getTitle()
    {
        return title;
    }

    /**
     * replaces old title
     * @return this ticket
     */
    public T setTitle(String title)
    {
        this.updatedFields.add(FieldNames.TITLE.name());
        this.title = title;

        logger.log(Level.FINEST, String.format(
            "[Ticket: %s] %s marked for update",
            this.id,
            FieldNames.TITLE.name()
        ));

        return (T) this;
    }

    /**
     * persist all fields marked as tampered into a request to update them in the real ticket system
     * @return the changed ticket requested from the real ticket system after updating
     */
    public abstract T save();

    @Override
    public boolean equals(Object o)
    {
        if (this == o) return true;
        if (o == null || !Objects.equals(o.getClass(), this.getClass())) return false;
        Ticket<?, ?> ticket = (Ticket<?, ?>) o;
        return Objects.equals(id, ticket.id) &&
            parent == ticket.parent;
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(id, parent);
    }

    @Override
    public String toString()
    {
        return "Ticket{" +
            "assignees: " + assignees +
            ", description='" + description + '\'' +
            ", id='" + id + '\'' +
            ", labels=" + labels +
            ", open=" + open +
            ", title='" + title + '\'' +
            '}';
    }
}
